# Copyright 2016 Infoblox Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""network_name_change_support

Revision ID: 0075c5a73439
Revises: 5267e17f370f
Create Date: 2016-08-09 16:19:00.207834

"""

# revision identifiers, used by Alembic.
revision = '0075c5a73439'
down_revision = '5267e17f370f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'infoblox_networks',
        sa.Column('network_id', sa.String(64),
                  nullable=False, primary_key=True),
        sa.Column('network_name', sa.String(255), nullable=False),
    )
